import base64
import datetime
import re
import struct
from html.entities import codepoint2name
from typing import Self
from xml.sax.saxutils import escape as xml_escape

from .constants import DICTIONARY
from .datatypes import Decimal, MultiByteInt31
from .record import record


# Regex to match invalid XML 1.0 characters (control chars except tab, newline, carriage return)
_INVALID_XML_CHARS = re.compile(r'[\x00-\x08\x0b\x0c\x0e-\x1f]')


def _sanitize_xml_text(value: str) -> str:
    """Sanitize a string for safe inclusion in XML.

    1. Removes invalid XML 1.0 control characters
    2. Escapes XML special characters (<, >, &, ", ')
    """
    # Remove invalid control characters
    clean = _INVALID_XML_CHARS.sub('', value)
    # Escape XML special characters
    return xml_escape(clean, entities={'"': '&quot;', "'": '&apos;'})


class Text(record): ...


class ZeroTextRecord(Text):
    type = 0x80

    def __str__(self):
        return "0"

    @classmethod
    def parse(cls, fp) -> Self:
        return cls()


class OneTextRecord(Text):
    type = 0x82

    def __str__(self):
        return "1"

    @classmethod
    def parse(cls, fp) -> Self:
        return cls()


class FalseTextRecord(Text):
    type = 0x84

    def __str__(self):
        return "false"

    @classmethod
    def parse(cls, fp) -> Self:
        return cls()


class TrueTextRecord(Text):
    type = 0x86

    def __str__(self):
        return "true"

    @classmethod
    def parse(cls, fp) -> Self:
        return cls()


class Int8TextRecord(Text):
    type = 0x88

    def __init__(self, value):
        self.value = value

    def to_bytes(self) -> bytes:
        return super(Int8TextRecord, self).to_bytes() + struct.pack("<b", self.value)

    def __str__(self):
        return str(self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        return cls(struct.unpack("<b", fp.read(1))[0])


class Int16TextRecord(Int8TextRecord):
    type = 0x8A

    def to_bytes(self) -> bytes:
        # print self.value
        return struct.pack("<B", self.type) + struct.pack("<h", self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        return cls(struct.unpack("<h", fp.read(2))[0])


class Int32TextRecord(Int8TextRecord):
    type = 0x8C

    def to_bytes(self) -> bytes:
        return struct.pack("<B", self.type) + struct.pack("<i", self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        return cls(struct.unpack("<i", fp.read(4))[0])


class Int64TextRecord(Int8TextRecord):
    type = 0x8E

    def to_bytes(self) -> bytes:
        return struct.pack("<B", self.type) + struct.pack("<q", self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        return cls(struct.unpack("<q", fp.read(8))[0])


class UInt64TextRecord(Int64TextRecord):
    type = 0xB2

    def to_bytes(self) -> bytes:
        return struct.pack("<B", self.type) + struct.pack("<Q", self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        return cls(struct.unpack("<Q", fp.read(8))[0])


class BoolTextRecord(Text):
    type = 0xB4

    def __init__(self, value):
        self.value = value

    def to_bytes(self) -> bytes:
        return struct.pack("<B", self.type) + struct.pack("<B", 1 if self.value else 0)

    def __str__(self):
        return str(self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        value = True if struct.unpack("<B", fp.read(1))[0] == 1 else False
        return cls(value)


class UnicodeChars8TextRecord(Text):
    type = 0xB6

    def __init__(self, string):
        self.value = string

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-16")[2:]  # skip bom
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<B", len(data))
        bytes += data
        return bytes

    def __str__(self):
        return _sanitize_xml_text(self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        ln: int = struct.unpack("<B", fp.read(1))[0]
        data: bytes = fp.read(ln)
        return cls(data.decode("utf-16"))


class UnicodeChars16TextRecord(UnicodeChars8TextRecord):
    type = 0xB8

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-16")[2:]  # skip bom
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<H", len(data))
        bytes += data
        return bytes

    def __str__(self):
        return _sanitize_xml_text(self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        ln: int = struct.unpack("<H", fp.read(2))[0]
        data: bytes = fp.read(ln)
        return cls(data.decode("utf-16"))


class UnicodeChars32TextRecord(UnicodeChars8TextRecord):
    type = 0xBA

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-16")[2:]  # skip bom
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<I", len(data))
        bytes += data
        return bytes

    def __str__(self):
        return _sanitize_xml_text(self.value)

    @classmethod
    def parse(cls, fp) -> Self:
        ln: int = struct.unpack("<I", fp.read(4))[0]
        data: bytes = fp.read(ln)
        return cls(data.decode("utf-16"))


class QNameDictionaryTextRecord(Text):
    type = 0xBC

    def __init__(self, prefix, index):
        self.prefix = prefix
        self.index = index

    def to_bytes(self):
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<B", ord(self.prefix) - ord("a"))
        bytes += MultiByteInt31(self.index).to_bytes()
        return bytes

    def __str__(self):
        return "%s:%s" % (self.prefix, DICTIONARY[self.index])

    @classmethod
    def parse(cls, fp) -> Self:
        prefix = chr(struct.unpack("<B", fp.read(1))[0] + ord("a"))
        index = MultiByteInt31.parse(fp).value
        return cls(prefix, index)


class FloatTextRecord(Text):
    type = 0x90

    def __init__(self, value):
        self.value: float = value

    def to_bytes(self) -> bytes:
        bytes = super(FloatTextRecord, self).to_bytes()
        bytes += struct.pack("<f", self.value)
        return bytes

    def __str__(self):
        try:
            if self.value == int(self.value):
                return "%.0f" % self.value
            else:
                return str(self.value)
        except:
            return str(self.value).upper()

    @classmethod
    def parse(cls, fp) -> Self:
        value = struct.unpack("<f", fp.read(4))[0]
        return cls(value)


class DoubleTextRecord(FloatTextRecord):
    type = 0x92

    def __init__(self, value: float):
        self.value = value

    def to_bytes(self) -> bytes:
        bytes = super(FloatTextRecord, self).to_bytes()
        bytes += struct.pack("<d", self.value)
        return bytes

    def __str__(self):
        return super(DoubleTextRecord, self).__str__()

    @classmethod
    def parse(cls, fp) -> Self:
        value = struct.unpack("<d", fp.read(8))[0]
        return cls(value)


class DecimalTextRecord(Text):
    type = 0x94

    def __init__(self, value: Decimal):
        self.value = value

    def __str__(self):
        return str(self.value)

    def to_bytes(self) -> bytes:
        return super(DecimalTextRecord, self).to_bytes() + self.value.to_bytes()

    @classmethod
    def parse(cls, fp) -> Self:
        value = Decimal.parse(fp)
        return cls(value)


class DatetimeTextRecord(Text):
    type = 0x96

    def __init__(self, value: int, tz: int):
        self.value = value
        self.tz = tz

    def __str__(self):
        ticks = self.value
        dt = datetime.datetime(1, 1, 1) + datetime.timedelta(microseconds=ticks / 10)
        return dt.isoformat()

    def to_bytes(self) -> bytes:
        bytes = super(DatetimeTextRecord, self).to_bytes()
        bytes += struct.pack(
            "<Q", (self.tz & 3) | (self.value & 0x1FFFFFFFFFFFFFFF) << 2
        )

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        data = struct.unpack("<Q", fp.read(8))[0]
        tz = data & 3
        value = data
        return cls(value, tz)


def escapecp(cp):
    return "&%s;" % codepoint2name[cp] if (cp in codepoint2name) else chr(cp)


def escape(text):
    newtext = ""
    for c in text:
        newtext += escapecp(ord(c))
    return newtext


class Chars8TextRecord(Text):
    type = 0x98

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return _sanitize_xml_text(self.value)

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-8")
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<B", len(data))
        bytes += data

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<B", fp.read(1))[0]
        value: str = fp.read(ln).decode("utf-8")
        return cls(value)


class Chars16TextRecord(Text):
    type = 0x9A

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return _sanitize_xml_text(self.value)

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-8")
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<H", len(data))
        bytes += data

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<H", fp.read(2))[0]
        value: str = fp.read(ln).decode("utf-8")
        return cls(value)


class Chars32TextRecord(Text):
    type = 0x9C

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return _sanitize_xml_text(self.value)

    def to_bytes(self) -> bytes:
        data = self.value.encode("utf-8")
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<I", len(data))
        bytes += data

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<I", fp.read(4))[0]
        value: str = fp.read(ln).decode("utf-8")
        return cls(value)


class UniqueIdTextRecord(Text):
    type = 0xAC

    def __init__(self, uuid):
        if isinstance(uuid, list) or isinstance(uuid, tuple):
            self.uuid = uuid
        else:
            if uuid.startswith("urn:uuid"):
                uuid = uuid[9:]
            uuid = uuid.split("-")
            tmp = uuid[0:3]
            tmp.append(uuid[3][0:2])
            tmp.append(uuid[3][2:])
            tmp.append(uuid[4][0:2])
            tmp.append(uuid[4][2:4])
            tmp.append(uuid[4][4:6])
            tmp.append(uuid[4][6:8])
            tmp.append(uuid[4][8:10])
            tmp.append(uuid[4][10:])

            self.uuid = [int(s, 16) for s in tmp]

    def to_bytes(self) -> bytes:
        bytes = super(UniqueIdTextRecord, self).to_bytes()
        bytes += struct.pack("<IHHBBBBBBBB", *self.uuid)

        return bytes

    def __str__(self):
        return "urn:uuid:%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % (
            tuple(self.uuid)
        )

    @classmethod
    def parse(cls, fp) -> Self:
        uuid = struct.unpack("<IHHBBBBBBBB", fp.read(16))
        return cls(uuid)


class UuidTextRecord(UniqueIdTextRecord):
    type = 0xB0

    def __str__(self):
        return "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % (tuple(self.uuid))


class Bytes8TextRecord(Text):
    type = 0x9E

    def __init__(self, data: bytes):
        self.value = data

    def to_bytes(self) -> bytes:
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<B", len(self.value))
        bytes += self.value

        return bytes

    def __str__(self):
        return base64.b64encode(self.value).decode()

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<B", fp.read(1))[0]
        data: bytes = struct.unpack("%ds" % ln, fp.read(ln))[0]
        return cls(data)


class Bytes16TextRecord(Text):
    type = 0xA0

    def __init__(self, data: bytes):
        self.value = data

    def __str__(self):
        return base64.b64encode(self.value).decode()

    def to_bytes(self) -> bytes:
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<H", len(self.value))
        bytes += self.value

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<H", fp.read(2))[0]
        data: bytes = struct.unpack("%ds" % ln, fp.read(ln))[0]
        return cls(data)


class Bytes32TextRecord(Text):
    type = 0xA2

    def __init__(self, data: bytes):
        self.value = data

    def __str__(self):
        return base64.b64encode(self.value).decode()

    def to_bytes(self) -> bytes:
        bytes = struct.pack("<B", self.type)
        bytes += struct.pack("<I", len(self.value))
        bytes += self.value

        return bytes

    @classmethod
    def parse(cls, fp) -> Self:
        ln = struct.unpack("<I", fp.read(4))[0]
        data: bytes = struct.unpack("%ds" % ln, fp.read(ln))[0]
        return cls(data)


class StartListTextRecord(Text):
    type = 0xA4


class EndListTextRecord(Text):
    type = 0xA6


class EmptyTextRecord(Text):
    type = 0xA8


class TimeSpanTextRecord(Text):
    type = 0xAE

    def __init__(self, value):
        self.value = value

    def to_bytes(self) -> bytes:
        return super(TimeSpanTextRecord, self).to_bytes() + struct.pack(
            "<q", self.value
        )

    def __str__(self):
        return str(datetime.timedelta(microseconds=self.value / 10))

    @classmethod
    def parse(cls, fp) -> Self:
        value = struct.unpack("<q", fp.read(8))[0]
        return cls(value)


class DictionaryTextRecord(Text):
    type = 0xAA

    def __init__(self, index):
        self.index = index

    def to_bytes(self) -> bytes:
        return (
            super(DictionaryTextRecord, self).to_bytes()
            + MultiByteInt31(self.index).to_bytes()
        )

    def __str__(self):
        return DICTIONARY[self.index]

    @classmethod
    def parse(cls, fp) -> Self:
        index = MultiByteInt31.parse(fp).value
        return cls(index)


record.add_records(
    (
        ZeroTextRecord,
        OneTextRecord,
        FalseTextRecord,
        TrueTextRecord,
        Int8TextRecord,
        Int16TextRecord,
        Int32TextRecord,
        Int64TextRecord,
        UInt64TextRecord,
        BoolTextRecord,
        UnicodeChars8TextRecord,
        UnicodeChars16TextRecord,
        UnicodeChars32TextRecord,
        QNameDictionaryTextRecord,
        FloatTextRecord,
        DoubleTextRecord,
        DecimalTextRecord,
        DatetimeTextRecord,
        Chars8TextRecord,
        Chars16TextRecord,
        Chars32TextRecord,
        UniqueIdTextRecord,
        UuidTextRecord,
        Bytes8TextRecord,
        Bytes16TextRecord,
        Bytes32TextRecord,
        StartListTextRecord,
        EndListTextRecord,
        EmptyTextRecord,
        TimeSpanTextRecord,
        DictionaryTextRecord,
    )
)



NAMESPACES = {
    "ad": "http://schemas.microsoft.com/2008/1/ActiveDirectory",
    "addata": "http://schemas.microsoft.com/2008/1/ActiveDirectory/Data",
    "adlq": "http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery",
    "da": "http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess",
    "ca": "http://schemas.microsoft.com/2008/1/ActiveDirectory/CustomActions",

    "soapenv": "http://www.w3.org/2003/05/soap-envelope",
    "wsa": "http://www.w3.org/2005/08/addressing",
    "wsen": "http://schemas.xmlsoap.org/ws/2004/09/enumeration",
    "wxf": "http://schemas.xmlsoap.org/ws/2004/09/transfer",
    "xsd": "http://www.w3.org/2001/XMLSchema",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",

    "s": "http://www.w3.org/2003/05/soap-envelope",
    "a": "http://www.w3.org/2005/08/addressing",
}


LDAP_ROOT_DSE_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<s:Header>
    <a:Action s:mustUnderstand="1" xmlns="http://www.w3.org/2005/08/addressing">http://schemas.xmlsoap.org/ws/2004/09/transfer/Get</a:Action>
    <ad:instance>ldap:389</ad:instance>
    <ad:objectReferenceProperty>11111111-1111-1111-1111-111111111111</ad:objectReferenceProperty>
    <a:MessageID>urn:uuid:{uuid}</a:MessageID>
    <a:ReplyTo>
        <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
    </a:ReplyTo>
    <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Resource</a:To>
</s:Header>
    <s:Body></s:Body>
</s:Envelope>"""


LDAP_QUERY_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <s:Header>
        <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate</a:Action>
        <ad:instance>ldap:389</ad:instance>
        <a:MessageID>urn:uuid:{uuid}</a:MessageID>
        <a:ReplyTo>
            <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
        </a:ReplyTo>
        <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Enumeration</a:To>
    </s:Header>
    <s:Body xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration"
        xmlns:adlq="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery">
        <wsen:Enumerate>
            <wsen:Filter Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/LdapQuery">
                <adlq:LdapQuery>
                    <adlq:Filter>{query}</adlq:Filter>
                    <adlq:BaseObject>{baseobj}</adlq:BaseObject>
                    <adlq:Scope>Subtree</adlq:Scope>
                </adlq:LdapQuery>
            </wsen:Filter>
            <ad:Selection Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/XPath-Level-1">
                {attributes}
            </ad:Selection>
        </wsen:Enumerate>
    </s:Body>
</s:Envelope>"""

LDAP_PULL_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
    xmlns:a="http://www.w3.org/2005/08/addressing"
    xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
    xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <s:Header>
        <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull</a:Action>
        <ad:instance>ldap:389</ad:instance>
        <a:MessageID>urn:uuid:{uuid}</a:MessageID>
        <a:ReplyTo>
            <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
        </a:ReplyTo>
        <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Enumeration</a:To>
    </s:Header>
    <s:Body xmlns:wsen="http://schemas.xmlsoap.org/ws/2004/09/enumeration">
        <wsen:Pull>
            <wsen:EnumerationContext>{enum_ctx}</wsen:EnumerationContext>
            <wsen:MaxElements>256</wsen:MaxElements>
                <ad:controls>
                    <ad:control criticality="true" type="1.2.840.113556.1.4.801">
                       <ad:controlValue xsi:type="xsd:base64Binary">MIQAAAADAgEH</ad:controlValue>
                </ad:control>
        </wsen:Pull>
    </s:Body>
</s:Envelope>"""


# Binary attributes that need base64 decoding
KNOWN_BINARY_ADWS_ATTRIBUTES = [
    "objectsid", "objectguid", "ntsecuritydescriptor", "sidhistory",
    "cacertificate", "pkiexpirationperiod", "pkioverlapperiod", 
    "msds-allowedtoactonbehalfofotheridentity", "msds-groupmsamembership"
]

LDAP_PUT_FSTRING: str = """<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
        xmlns:a="http://www.w3.org/2005/08/addressing"
        xmlns:addata="http://schemas.microsoft.com/2008/1/ActiveDirectory/Data"
        xmlns:ad="http://schemas.microsoft.com/2008/1/ActiveDirectory"
        xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xmlns:da="http://schemas.microsoft.com/2006/11/IdentityManagement/DirectoryAccess">
        <s:Header>
            <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2004/09/transfer/Put</a:Action>
            <ad:instance>ldap:389</ad:instance>
            <ad:objectReferenceProperty>{object_ref}</ad:objectReferenceProperty>
            <da:IdentityManagementOperation s:mustUnderstand="1"
                xmlns:i="http://www.w3.org/2001/XMLSchema-instance"></da:IdentityManagementOperation>
            <a:MessageID>urn:uuid:{uuid}</a:MessageID>
            <a:ReplyTo>
                <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
            </a:ReplyTo>
            <a:To s:mustUnderstand="1">net.tcp://{fqdn}:9389/ActiveDirectoryWebServices/Windows/Resource</a:To>
        </s:Header>
        <s:Body>
            <da:ModifyRequest Dialect="http://schemas.microsoft.com/2008/1/ActiveDirectory/Dialect/XPath-Level-1">
                <da:Change Operation="{operation}">
                    <da:AttributeType>{attribute}</da:AttributeType>
                    <da:AttributeValue>
                        <ad:value xsi:type="xsd:{data_type}">{value}</ad:value>
                    </da:AttributeValue>
                </da:Change>
            </da:ModifyRequest>
        </s:Body>
    </s:Envelope>"""



"""
ADWS wrapper to replace LDAP functionality
This module provides a compatibility layer that mimics ldap3's Connection and Entry classes
but uses ADWS (Active Directory Web Services) instead of LDAP.
"""
import sys
import os
import base64
from xml.etree import ElementTree
from uuid import UUID
from datetime import datetime
from typing import List, Optional, Any
import logging

# Use local standalone ADWS implementation
try:
    from .adws import ADWSConnect, NTLMAuth, ADWSError, NAMESPACES, KNOWN_BINARY_ADWS_ATTRIBUTES
    from impacket.ldap.ldaptypes import LDAP_SID
except ImportError as e:
    raise ImportError(f"Failed to import ADWS modules: {e}")

logger = logging.getLogger(__name__)


class ADWSEntry:
    """
    Mimics ldap3 Entry class for compatibility with existing code
    """
    def __init__(self, xml_element: ElementTree.Element, attributes: List[str]):
        self._xml_element = xml_element
        self._attributes = {}
        self._parse_xml_element(attributes)
    
    def _parse_xml_element(self, requested_attributes: List[str]):
        """Parse XML element and extract attribute values"""
        for attr_name in requested_attributes:
            attr_name_lower = attr_name.lower()
            # Try original case first
            attr_elems = self._xml_element.findall(f".//addata:{attr_name}/ad:value", namespaces=NAMESPACES)
            if not attr_elems and attr_name != attr_name_lower:
                # Try lowercase
                attr_elems = self._xml_element.findall(f".//addata:{attr_name_lower}/ad:value", namespaces=NAMESPACES)
            
            if attr_elems:
                values = []
                for val_elem in attr_elems:
                    if val_elem.text is None:
                        continue
                    is_b64_by_type = val_elem.attrib.get('{http://www.w3.org/2001/XMLSchema-instance}type') == 'ad:base64Binary'
                    if is_b64_by_type or attr_name_lower in KNOWN_BINARY_ADWS_ATTRIBUTES:
                        if isinstance(val_elem.text, str):
                            try:
                                values.append(base64.b64decode(val_elem.text))
                            except Exception:
                                values.append(val_elem.text)
                        else:
                            values.append(val_elem.text)
                    else:
                        values.append(val_elem.text)
                
                if values:
                    # Store as single value if only one, otherwise as list
                    if len(values) == 1:
                        self._attributes[attr_name] = ADWSAttribute(attr_name, values[0])
                    else:
                        self._attributes[attr_name] = ADWSAttribute(attr_name, values)
        
        # Always try to get distinguishedName and objectClass
        if 'distinguishedName' not in self._attributes:
            dn_elem = self._xml_element.find(".//addata:distinguishedName/ad:value", namespaces=NAMESPACES)
            if dn_elem is not None and dn_elem.text is not None:
                self._attributes['distinguishedName'] = ADWSAttribute('distinguishedName', dn_elem.text)
        
        if 'objectClass' not in self._attributes:
            oc_vals = [oc.text for oc in self._xml_element.findall(".//addata:objectClass/ad:value", namespaces=NAMESPACES) if oc.text]
            if oc_vals:
                self._attributes['objectClass'] = ADWSAttribute('objectClass', oc_vals if len(oc_vals) > 1 else oc_vals[0])
    
    def __getitem__(self, key):
        """Allow dictionary-like access"""
        if key in self._attributes:
            return self._attributes[key]
        raise KeyError(f"Attribute '{key}' not found")
    
    def __contains__(self, key):
        """Check if attribute exists"""
        return key in self._attributes
    
    def get(self, key, default=None):
        """Get attribute with default"""
        return self._attributes.get(key, default)
    
    def entry_to_json(self):
        """Convert entry to JSON string (mimics ldap3)"""
        import json
        result = {}
        for key, attr in self._attributes.items():
            if isinstance(attr, ADWSAttribute):
                result[key] = attr.value if not isinstance(attr.value, bytes) else base64.b64encode(attr.value).decode('ascii')
            else:
                result[key] = attr
        return json.dumps(result)


class ADWSAttribute:
    """
    Mimics ldap3 Attribute class
    """
    def __init__(self, key: str, value: Any):
        self.key = key
        self._value = value
        self._raw_values = [value] if not isinstance(value, list) else value
    
    @property
    def value(self):
        """Get single value (first if list)"""
        if isinstance(self._value, list):
            return self._value[0] if self._value else None
        return self._value
    
    @property
    def values(self):
        """Get all values as list"""
        if isinstance(self._value, list):
            return self._value
        return [self._value]
    
    @property
    def raw_values(self):
        """Get raw values"""
        return self._raw_values


class ADWSServer:
    """
    Mimics ldap3 Server class for compatibility
    """
    def __init__(self, host: str, domain: str):
        self.host = host
        self.domain = domain
        self.info = ADWSServerInfo(domain)
    
    def __repr__(self):
        return f"ADWSServer(host={self.host}, domain={self.domain})"


class ADWSServerInfo:
    """
    Mimics ldap3 Server.info for compatibility
    """
    def __init__(self, domain: str):
        self.other = {
            'defaultNamingContext': [f"DC={',DC='.join(domain.split('.'))}"]
        }


class ADWSConnection:
    """
    Mimics ldap3 Connection class but uses ADWS
    """
    def __init__(self, server: ADWSServer, user: Optional[str] = None, 
                 password: Optional[str] = None, authentication: Any = None):
        self.server = server
        self.user = user
        self.password = password
        self.authentication = authentication
        self.entries = []
        self._adws_client = None
        self._root_dn = None
        self._bound = False
    
    def bind(self) -> bool:
        """Bind to ADWS server"""
        try:
            # Parse domain and username
            domain = self.server.domain
            username = self.user or ""
            
            if '\\' in username:
                domain, username = username.split('\\', 1)
            elif '@' in username:
                username, domain = username.rsplit('@', 1)
            
            # Determine if password is a hash (LM:NTLM format)
            hashes = None
            password = self.password
            if password and ':' in password and len(password.split(':')) == 2:
                hashes = password
                password = None
            
            # Create NTLM auth
            auth = NTLMAuth(password=password, hashes=hashes)
            
            # Create ADWS connection
            self._adws_client = ADWSConnect.pull_client(
                ip=self.server.host,
                domain=domain,
                username=username,
                auth=auth
            )
            
            # Get root DN from RootDSE
            rootdse = self._adws_client.get_rootdse_contexts(self.server.host, self._adws_client._nmf)
            self._root_dn = rootdse.get('defaultNamingContext') or rootdse.get('rootDomainNamingContext')
            if self._root_dn:
                self.server.info.other['defaultNamingContext'] = [self._root_dn]
            
            self._bound = True
            return True
        except Exception as e:
            logger.error(f"ADWS bind failed: {e}")
            self.result = {'description': str(e)}
            return False
    
    def search(self, search_base: str, search_filter: str, attributes: List[str] = None):
        """
        Search using ADWS (mimics ldap3 search)
        """
        if not self._bound:
            raise RuntimeError("Not bound to ADWS server")
        
        if attributes is None:
            attributes = ['*']
        
        # Use ALL_ATTRIBUTES equivalent
        if attributes == ['*'] or (isinstance(attributes, list) and len(attributes) == 1 and attributes[0] == '*'):
            # For now, use a comprehensive list - in production you might want to query schema
            attributes = [
                'cn', 'name', 'sAMAccountName', 'distinguishedName', 'objectClass',
                'objectSid', 'objectGUID', 'memberOf', 'primaryGroupId', 'userAccountControl',
                'whenCreated', 'whenChanged', 'lastLogon', 'pwdLastSet', 'description',
                'servicePrincipalName', 'operatingSystem', 'operatingSystemServicePack',
                'operatingSystemVersion', 'dNSHostName', 'lockOutObservationWindow',
                'lockoutDuration', 'lockoutThreshold', 'maxPwdAge', 'minPwdAge',
                'minPwdLength', 'pwdHistoryLength', 'pwdProperties',
                'ms-DS-MachineAccountQuota', 'flatName', 'securityIdentifier',
                'trustAttributes', 'trustDirection', 'trustType', 'member'
            ]
        else:
            # Always include distinguishedName if not already present
            if 'distinguishedName' not in attributes:
                attributes = attributes + ['distinguishedName']
        
        try:
            # Use pull to get all results
            pull_result = self._adws_client.pull(
                query=search_filter,
                attributes=attributes,
                base_object_dn_for_soap=search_base
            )
            
            if pull_result is None:
                self.entries = []
                return
            
            # Convert XML elements to ADWSEntry objects
            self.entries = []
            for item_elem in pull_result:
                entry = ADWSEntry(item_elem, attributes)
                if 'distinguishedName' in entry:
                    self.entries.append(entry)
        except Exception as e:
            logger.error(f"ADWS search failed: {e}")
            self.entries = []
    
    class ExtendStandard:
        """Mimics ldap3 extend.standard for paged_search"""
        def __init__(self, connection):
            self._connection = connection
        
        def paged_search(self, search_base: str, search_filter: str, 
                         attributes: List[str] = None, paged_size: int = 500, 
                         generator: bool = False):
            """
            Paged search (ADWS handles paging internally, so this is just a regular search)
            """
            self._connection.search(search_base, search_filter, attributes)
            if generator:
                return iter(self._connection.entries)
            return self._connection.entries
    
    class Extend:
        """Mimics ldap3 extend for paged_search"""
        def __init__(self, connection):
            self._connection = connection
        
        @property
        def standard(self):
            """Mimics ldap3 extend.standard"""
            return ADWSConnection.ExtendStandard(self._connection)
    
    @property
    def extend(self):
        """Mimics ldap3 extend property"""
        return ADWSConnection.Extend(self)




import logging as log

from .record import Element, EndElementRecord, record
from .text import Text


def print_records(records, first_call=True) -> str:
    """
    returns the given record tree as a string
    """

    if not records:
        return ""

    output = ""
    for r in records:
        if isinstance(r, EndElementRecord):
            continue

        output += str(r)
        new_line = ""
        if hasattr(r, "childs"):
            new_line = print_records(r.childs, False)
        if isinstance(r, Element):
            output += new_line
            if hasattr(r, "prefix"):
                output += "</%s:%s>" % (r.prefix, r.name)
            else:
                output += "</%s>" % r.name
    return output


def pretty_print_records(records, skip=0, first_call=True) -> str:
    """prints the given record tree into a file like object"""

    if not records:
        return ""

    output = ""
    for r in records:
        if isinstance(r, EndElementRecord):
            continue
        if isinstance(r, Element):
            output += str(r)
        else:
            output += str(r)

        new_line = ""
        if hasattr(r, "childs"):
            new_line = pretty_print_records(r.childs, skip + 1, False)
        if isinstance(r, Element):
            output += new_line
            if new_line:
                output += "\r\n" + " " * skip
            if hasattr(r, "prefix"):
                output += "</%s:%s>" % (r.prefix, r.name)
            else:
                output += "</%s>" % r.name
    return output


def dump_records(records: list[record]) -> bytes:
    """
    returns the byte representation of a given record tree

    """
    out = b""

    for r in records:
        msg = f"Write {type(r).__name__}"

        if r == records[-1] and isinstance(r, Text):
            r.type = r.type + 1
            msg += " with EndElement (0x%X)" % r.type
        log.debug(msg)
        log.debug(f"Value {r}")

        if (
            isinstance(r, Element)
            and not isinstance(r, EndElementRecord)
            and len(r.attributes)
        ):
            log.debug(" Attributes:")
            for a in r.attributes:
                log.debug(f" {type(a).__name__}: {a}")

        out += r.to_bytes()

        if hasattr(r, "childs"):
            out += dump_records(r.childs)

            # only print the end element if the current record is NOT a "*WithEnd" record type
            if (not r.childs or not isinstance(r.childs[-1], Text)) and not isinstance(
                r, Text
            ):
                log.debug(f"Write EndElement for {type(r).__name__}")
                out += EndElementRecord().to_bytes()

        elif isinstance(r, Element) and not isinstance(r, EndElementRecord):
            log.debug(f"Write EndElement for {type(r).__name__}")
            out += EndElementRecord().to_bytes()

    return out


class Net7BitInteger(object):
    @staticmethod
    def decode7bit(data: bytes) -> tuple[int, int]:
        MAXMBI = 0x7F

        length = 0
        value = 0
        for i in range(5):
            length += 1
            v = data[i]
            value |= (v & MAXMBI) << 7 * i
            if not v & (MAXMBI + 1):
                break
        return value, length

    @staticmethod
    def encode7bit(value):
        MAXMBI = 0x7F
        if value < 0:
            raise ValueError("Signed numbers are not supported")

        if value <= MAXMBI:
            return bytes([value])

        result = []
        for _ in range(5):
            byte = value & MAXMBI
            value >>= 7
            if value != 0:
                byte |= MAXMBI + 1
            result.append(byte)

            if value == 0:
                break

        return bytes(result)
